/*-----------------------------------------------------------------------------
 *
 * Enhydra JdbcS
 * The Initial Developer of the Original Code is Lutris Technologies Inc.
 * Portions created by Lutris are Copyright 1997-2000 Lutris Technologies Inc.
 * All Rights Reserved.
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 * www.enhydra.org/license/epl.html
 *
 * Software distributed under the License is distributed on an "ASIS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 * ----------------------------------------------------------------------------
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * -----------------------------------------------------------------------------
 */
package org.enhydra.jdbc.pool;

import java.sql.SQLException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.StringRefAddr;
import javax.sql.ConnectionEvent;
import javax.sql.XAConnection;
import javax.sql.XADataSource;
import javax.transaction.TransactionManager;
import javax.transaction.Transaction;
import javax.transaction.SystemException;
import javax.transaction.xa.XAResource;
import org.enhydra.jdbc.standard.StandardXADataSource;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * StandardXAPoolDataSource class allows to make some operations on
 * XAConnection. It implements PoolHelper for the 3 methods :<p>
 * create   : create an XAConnection<p>
 * create(user,password)   : create a PooledConnection with an other user/password<p>
 * testThisObject : check if the object is still valid<p>
 * checkThisObject : check if the object is closed<p>
 * expire   : kill the object<p>
 */
public class StandardXAPoolDataSource extends StandardPoolDataSource {

    public XADataSource 	xads;		// object to build XAConnection
    public TransactionManager 	transactionManager; // the current Transaction Manager
    public Log glog = LogFactory.getLog("xapool");

    /**
     * Constructor
     */
    public StandardXAPoolDataSource() {
        super();
    }

    /**
     * Constructor
     */
    public StandardXAPoolDataSource(int initSize) {
        super(initSize);
    }

    /**
     * Constructor
     */
    public StandardXAPoolDataSource(StandardXADataSource ds) {
        super(ds);
    }

    /**
     * Constructor
     */
    public StandardXAPoolDataSource(StandardXADataSource ds, int initSize) {
        super(ds, initSize);
    }

    public void setTransactionManager(TransactionManager tm) {
        transactionManager = tm;
    }

    public void initContext() throws SQLException {
        if (xads == null) {
            log.debug("StandardXAPoolDataSource:initContext");
            try {
                ictx =  new InitialContext();		// get the initial context
                xads = (XADataSource)ictx.lookup(this.dataSourceName);// lookup the symbol to create object
                log.debug("StandardXAPoolDataSource:initContext lookup done");
            } catch (javax.naming.NameNotFoundException e) {
                log.error("StandardXAPoolDataSource:initContext The XADataSource is not bound at " + this.dataSourceName);
                throw new SQLException("The XADataSource is not bound at " + this.dataSourceName);
            } catch(javax.naming.NamingException e) {
                log.error("StandardXAPoolDataSource:initContext Error javax.naming.NamingException in StandardXAPoolDataSource:create");
            }
        }
    }

    /**
     * Invoked when the application calls close()
     * on its representation of the connection
     */
    public void connectionClosed(ConnectionEvent event) {
        Object obj = event.getSource();
        log.debug("StandardXAPoolDataSource:connectionClosed");
        XAConnection xac = (XAConnection)obj;	// cast it into an xaConnection

        Transaction tx = null;
        try {
            tx = transactionManager.getTransaction();
            log.debug("StandardXAPoolDataSource:connectionClosed get a transaction");
        } catch (NullPointerException n) {
            // current is null: we are not in EJBServer.
            log.error("StandardXAPoolDataSource:connectionClosed should not be used outside an EJBServer");
        } catch (SystemException e) {
            log.error("StandardXAPoolDataSource:connectionClosed getTransaction failed:"+e);
        }

        // delist Resource if in transaction
        // We must keep the connection till commit or rollback
        if (tx != null) {
            try {
                tx.delistResource(xac.getXAResource(), XAResource.TMSUCCESS);	// delist the xaResource
            } catch (Exception e) {
                log.error("StandardXAPoolDataSource:connectionClosed Exception in connectionClosed:"+e);
            }
        }
        log.debug("StandardXAPoolDataSource:connectionClosed checkIn an object to the pool");
        pool.checkIn(obj);	// return the connection to the pool
    }

    public GenerationObject create(String _user, String _password) throws SQLException{
        GenerationObject genObject;
        XAConnection xaCon =
                xads.getXAConnection(_user,_password);       // get the xa connection
        xaCon.addConnectionEventListener(this);         // add it to the event listener
        log.debug("StandardXAPoolDataSource:create create a object for the pool");
        genObject = new GenerationObject(xaCon, pool.getGeneration(), _user, _password);

        return genObject;
    }

    /**
     * Retrieves the Reference of this object. Used at binding time by JNDI
     * to build a reference on this object.
     *
     * @return  The non-null Reference of this object.
     * @exception  NamingException  If a naming exception was encountered while
     * retrieving the reference.
     */
    public Reference getReference() throws NamingException {
        log.debug("StandardXAPoolDataSource:getReference return a reference of the object");
        //Reference ref = new Reference(getClass().getName(), getClass().getName(), null);
        Reference ref = super.getReference();
        ref.add(new StringRefAddr("checkLevelObject", Integer.toString(getCheckLevelObject())));
        ref.add(new StringRefAddr("lifeTime", Long.toString(getLifeTime())));
        ref.add(new StringRefAddr("jdbcTestStmt", getJdbcTestStmt()));
        ref.add(new StringRefAddr("maxSize", Integer.toString(getMaxSize())));
        ref.add(new StringRefAddr("minSize", Integer.toString(getMinSize())));
        ref.add(new StringRefAddr("dataSourceName", getDataSourceName()));

        return ref;
    }

    /** Getter for property dataSource.
     * @return Value of property dataSource.
     */
    public XADataSource getDataSource() {
        return xads;
    }

    /** Setter for property dataSource.
     * @param dataSource New value of property dataSource.
     */
    public void setDataSource(XADataSource dataSource) {
        this.xads = dataSource;
        transactionManager = ((StandardXADataSource)dataSource).transactionManager;
    }
}
